#!BPY

""" Registration info for Blender menus:
Name: 'CD++ Simulations'
Blender: 241
Group: 'Simulations'
Tooltip: 'CD++ Simulations2'
"""

# --------------------------------------------------------------------------
# ***** BEGIN GPL LICENSE BLOCK *****
#
# Copyright (C) 2007 Emil Poliakov aka NiTeC
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# 
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------

__author__ = "Emil Poliakov"
__version__ = "1.0.2"

# --------------------------------------------------------------------------        
# Modified by Kin Wing Tsui - Dec 2007
# To animate the auto factory DEVS model
#
# Modified by Patrick Castonguay - 4 Sept 2008
# To animate the aircraft evacuation model
#   Fixed file openning and closing to prevent buffer overrun
#   Added the parsing of .ma and .val files
#   Added the logging of information to a file
# --------------------------------------------------------------------------

__bpydoc__ = """
CD++ Simulations 1.0.2

With this script you can load the CD++ Simulations results into Blender and apply them to objects etc.

Select the proper .MA and .LOG file via the graphical interface

If refence to a .val file is made, the file is also parsed and initial values used to set visualisation
"""

import Blender  # This will import the library of blender functions we will use
#import bpy
from Blender import Draw, BGL, Window, Image, Scene

winsize = Window.GetAreaSize()
scn = Scene.GetCurrent()
#scn = bpy.data.scenes.active
# myimage = Image.Load('c:\cd.jpg')

defaultPath = "c:/Program Files (x86)/Blender Foundation/Blender/.Blender/"
logFileName = defaultPath+"ac_evac20LOG.log"
maFileName = defaultPath+"ac_evac20.ma"
valfilename = ""
defaultValue = 0

# --------------------------------------------------------------------------
# read_val(path)
#
# Read and parse CD++ val file
#   Setup initial values for visual environment
#
# path : full path of the CD++ val file
# --------------------------------------------------------------------------
def read_val(path):
        global logfile
        logfile.write("\n\n***************\nReading file "+path+"\n")
        
	Blender.Window.WaitCursor(1)
	try:
            valFile = open(path, "r")
	except:
            logfile.write("    Error in read_val("+path+")\n")
            return
	
	# Look for cell information
	for line in valFile.readlines():
		words = line.split('=')
		# Will be true if '=' is found on line
		if len(words) >= 2:
			time          = "00:00:00:000"
			cell          = words[0]
			logValueWorld = words[1]
			try:
                            apply_log(cell.strip(), time, logValueWorld.strip())
			except:
                            valFile.close()
                            cdppLogFile.close()
        valFile.close()
	Blender.Window.WaitCursor(0)
	
# --------------------------------------------------------------------------
# read_ma(path)
#
# Read and parse CD++ ma file
#   Currently only looks for the .val file to setup visual environment
#
# path : full path of the CD++ ma file
# --------------------------------------------------------------------------
def read_ma(path):
	global logfile, valfilename, defaultvalue
	logfile.write("\n\n***************\nReading file "+path+"\n")
        
	Blender.Window.WaitCursor(1)
	try:
            maFile = open(path, "r")
	except ValueError:
            logfile.write("    Error in read_ma("+path+")\n")
            return
	
	# Look for a .val file and defaultInitialValue
	for line in maFile.readlines():
                words = line.split()
                # Keywords are expected to be first word of line
                if len(words) > 0:
                    if words[0] == ("initialvalue"):
                        defaultValue = words[2]
                    elif words[0] == ("initialCellsValue"):
                        valfilename = defaultPath+words[2]
                        read_val(valfilename)
        maFile.close()
	Blender.Window.WaitCursor(0)
	
# --------------------------------------------------------------------------
# read_log(path)
#
# Read and parse CD++ log file
#
# path : full path of the CD++ log file
# --------------------------------------------------------------------------
def read_log(path):
        global logfile
        logfile.write("\n\n***************\nReading file "+path+"\n")
        
	Blender.Window.WaitCursor(1)
	try:
            cdppLogFile = open(path, "r")
	except ValueError:
            logfile.write("    Error in read_log("+path+")\n")
            return

        # Look for messages in the CD++ log file
	for line in cdppLogFile.readlines():
		words = line.split()
		# For ac_eval only Y messages are processed
		if (len(words) > 1) and (words[1] == ("Y")):
			time          = words[3]
			cell          = words[5]
			#port          = words[7]
			logValueWorld = words[9]
			try:
                            apply_log(cell.strip(), time, logValueWorld.strip())
                        except:
                            cdppLogFile.close()
	cdppLogFile.close()
	Blender.Window.WaitCursor(0)

# --------------------------------------------------------------------------
# apply_log(cell, time, logValueWord)
#
# Interpret CD++ log file data for Blender animation
#
# cell :        cell to handle 
# time :	timestamp of the message
# valueWord :	string containing the numeric value in the message
# --------------------------------------------------------------------------
def apply_log(cell, time, logValueWorld):
        global logfile
        logfile.write("Simulation time: "+time+", Processing "+cell+" with value "+logValueWorld+"\n")

        # Time expected to be CD++ kind 00:00:00:000; transform it for Blender
        # It is used as a frame setup.  For ac_evac the increment used is seconds
	hours = (time[0:2])
	minutes = (time[3:5])
	seconds = (time[6:8])
	mseconds = (time[9:12])
#	totaltime = (int(hours)*3600 + int(minutes)*60 + int(seconds)*10 + int(mseconds)) #old file
	totaltime = (int(hours)*3600 + int(minutes)*60 + int(seconds))
	Blender.Set('curframe',totaltime)
	
        # Get cell description currently works for three dimensions
        # cells are always described as 'modelName'(x,y,z)(..)
        # zeroize position counters to be more versatile than previous version
        a = b = c = d = 0
	for i in range(len(cell)):
		if cell[i]=='(' and cell[i-1]!=')':
			a = i
		elif cell[i]==',' and b==0:
			b = i
		elif cell[i]==',':
			c = i
		elif cell[i]==')' and d==0:
			d = i
			
        # Transform description into coordinate			
	xcoord = (cell[a+1:(a+len(range(a,b)))])
	ycoord = (cell[b+1:(b+len(range(b,c)))])
	zcoord = (cell[c+1:(c+len(range(c,d)))])
        
        # Only first plane is used in ac_evac simulation to represent movement
        # the other is used for background calculation within the CD++ model
        if int(zcoord) != 0:
 		return #Exit the function apply_log()

        # Here we truncate the decimal protion as the value is expected to be an integer  
        logend = len(logValueWorld)
	for j in range(len(logValueWorld)):
		if logValueWorld[j]=='.':
			logend = j
	logValue = int(logValueWorld[0:logend])

        # Set object name to the current cell
	objectname = ("Cell_"+xcoord+"_"+ycoord+"_"+zcoord)
	logfile.write("    Processing: "+objectname+"\n")

        scene_obs = list(scn.objects)
        try:
            # Process state values that are passed as logValue for the current cell	
            if (logValue == -2): # Occupied cell (passenger)
		try:
                        #ob = [o for o in scn.objects if o.name == objectname][0]
			ob = Blender.Object.Get(objectname)
			# If object not already linked to current scene
			# this is done as ac_evac had duplicates of object creation
                        if (ob not in scene_obs):
                            scn.objects.link(ob)
                            logfile.write("    Object already exist but has been relinked\n")
                        else:
                            logfile.write("    Object already exist and is already linked\n")
		except ValueError:
                        #human = [o for o in scn.objects if o.name == "Human"][0]
			human = Blender.Object.Get ('Human')
			human.select(1)
			Blender.Object.Duplicate()
			activeObject = scn.getActiveObject()
			activeObject.name = objectname
			activeObject.LocX = int(xcoord)
			activeObject.LocY = int(ycoord)
			activeObject.LocZ = int(zcoord)
			activeObject.select(0)
			logfile.write("    Object created\n")
            elif (logValue == -1): # Wall or obstacle
		try:
                        #cube = [o for o in scn.objects if o.name == "Cube"][0]
                        cube = Blender.Object.Get('Cube')
			cube.select(1)
			Blender.Object.Duplicate()
			activeObject = scn.getActiveObject()
			activeObject.name = objectname
			activeObject.LocX = int(xcoord)
			activeObject.LocY = int(ycoord)
			activeObject.LocZ = int(zcoord)
			activeObject.select(0)
			logfile.write("    Object cube created\n")
		except ValueError:
			logfile.write("    **Object cube already exist or error occured**\n")
            elif (logValue == -9): # Chair
		try:
                        #TODO change to chair after it is imported to ac_evac.blend
                        #cube = [o for o in scn.objects if o.name == "Cube"][0]
                        cube = Blender.Object.Get('Chair')
			cube.select(1)
			Blender.Object.Duplicate()
			activeObject = scn.getActiveObject()
			activeObject.name = objectname
			activeObject.LocX = int(xcoord)
			activeObject.LocY = int(ycoord)
			activeObject.LocZ = int(zcoord)
			activeObject.select(0)
			logfile.write("    Object chair created\n")
		except ValueError:
			logfile.write("    **Object chair already exist or error occured**\n")
            elif (logValue == 9): # Emergency exit
		try:
                        #cubeExit = [o for o in scn.objects if o.name == "Cube.Exit"][0]
			cubeExit = Blender.Object.Get('Cube.Exit')
			cubeExit.select(1)
			Blender.Object.Duplicate()
			activeObject = scn.getActiveObject()
			activeObject.name = objectname
			activeObject.LocX = int(xcoord)
			activeObject.LocY = int(ycoord)
			activeObject.LocZ = int(zcoord)
			activeObject.select(0)
		except ValueError:
			logfile.write("    **Object cube already exist or error occured**\n")
            elif (logValue == 0): # Empty cell
		try:
                        #ob = [o for o in scn.objects if o.name == objectname][0]
			ob = Blender.Object.Get(objectname)
			scn.objects.unlink(ob)
			logfile.write("    Object unlinked\n")
		except ValueError:
			logfile.write("    Cell already empty\n")
            else:
                logfile.write("    **Log value not recognised as valid: %d**\n" %(logValue))

        except ValueError:
            logfile.write("\n\n\n")
            logfile.write("****************************************\n")
            logfile.write("* Unexpected stop to logging segence   *\n")
            logfile.write("*   in apply_log() portion of script   *\n")            
            logfile.write("****************************************\n")
            logfile.close()
            
        # Refresh the display
        scn.update()
	Blender.Redraw()
#	print 'The current frame is:' , Blender.Get('curframe'), " !!!END OF SCRIPT!!!"

# --------------------------------------------------------------------------
# import_maFile(ma)
#
# Saves the CD++ MA file path
#
# ma : MA file path to be saved
# --------------------------------------------------------------------------    
def import_maFile(ma):
	global maFileName
	maFileName = ma
	Draw.Redraw(1)

# --------------------------------------------------------------------------
# import_logFile(log)
#
# Saves the CD++ log file path
#
# log : log file path to be saved
# --------------------------------------------------------------------------    
def import_logFile(log):
	global logFileName
	logFileName = log
	Draw.Redraw(1)

# --------------------------------------------------------------------------
# event(evt, val)
#
# Keyboard and mouse event handler
# -------------------------------------------------------------------------- 
def event(evt, val):  
     return

# --------------------------------------------------------------------------
# buttonHandler(evt)
#
# Button-click handler
# Handles .ma and .log file Browse Buttons as well as Execute Button
# --------------------------------------------------------------------------
def buttons(evt): 
    global logFileName, maFileName, logfile
    default_name = Blender.Get('filename')
   # if default_name.endswith('untitled.blend'):
   # Could add handling for file path default setup
    ma_name = default_name[0:len(default_name)-len("untitled.blend")] + ('*.ma')
    log_name = default_name[0:len(default_name)-len("untitled.blend")] + ('*.log')
    if evt == 1:
        Blender.Window.FileSelector(import_maFile, "Select",ma_name)
    if evt == 2:
        Blender.Window.FileSelector(import_logFile, "Select", log_name)
    if evt == 3:
        #TODO disable Execute button
        
        # Set file where debuging and execution informations will be saved
        logfile = open(defaultPath+"cdpp_log.txt","w")
        logfile.write("****************************\n")
        logfile.write("* Start of logging segence *\n")
        logfile.write("****************************\n\n\n")
        print "Executing CD++/Blender Simulation Visualisation Tool..."
        
        read_ma(maFileName)
        read_log(logFileName)
        
        print "Execution complete, see log file for details:\n   ", logfile.name
        logfile.write("\n\n\n")
        logfile.write("****************************\n")
        logfile.write("* End of logging segence   *\n")
        logfile.write("****************************\n")
        logfile.close()

        BGL.glClearColor(0,0,0,1)
        BGL.glClear(Blender.BGL.GL_COLOR_BUFFER_BIT)

# --------------------------------------------------------------------------
# gui()
#
# Main initial user interface of the script
# Sets-up the display
# --------------------------------------------------------------------------        
def gui():
    global maFileName, logFileName, loginfo
    BGL.glClearColor(0,0,0,1)
    BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
    BGL.glColor3f(1,1,1) 

    # MA Browse button setup  
    BGL.glRasterPos2i(10, (winsize[1]-100))
    text1 = Draw.Text("Select MA File", 'normal')
    log_toggle = Draw.Toggle("Browse",1,10,(winsize[1]-130),100,20,0,"Import the CD++ MA File")
    BGL.glRasterPos2i(10, (winsize[1]-160))
    text1 = Draw.Text(maFileName, 'normal')

    # LOG Browse button setup
    BGL.glRasterPos2i(10, (winsize[1]-200))
    text2 = Draw.Text("Select LOG File", 'normal')
    ma_toggle = Draw.Toggle("Browse",2,10,(winsize[1]-230),100,20,0,"Import the CD++ Generated LOG File")
    BGL.glRasterPos2i(10, (winsize[1]-260))
    text1 = Draw.Text(logFileName, 'normal')

    # Execute button setup
    execute_toggle = Draw.Toggle("Execute",3,10,(winsize[1]-350),100,20,0,"Starts the Blender Visualization")
 
    # Text at the top of the screen
    BGL.glRasterPos2i(10, (winsize[1]-20))
    Draw.Text("Welcome to the CD++/Blender", 'large')
    BGL.glRasterPos2i(10, (winsize[1]-40))
    Draw.Text("Simulation Visualization Tool", 'large')

    # Display image on the script screen
    #BGL.glBlendFunc(BGL.GL_SRC_ALPHA, BGL.GL_ONE_MINUS_SRC_ALPHA) 
    #Draw.Image(myimage, 10, winsize[1]-70)

# Register user interface and user event handlers
Blender.Draw.Register(gui, event, buttons)

#!BPY


	
